"""An opionionated wrapper around :class:`pyg4ometry.visualization.VtkViewerNew`."""

from __future__ import annotations

import argparse
import copy
import logging
import re
from pathlib import Path

import awkward as ak
import numpy as np
import pyg4ometry.geant4 as g4
import vtk
from dbetto.utils import load_dict
from pyg4ometry import config as meshconfig
from pyg4ometry import gdml
from pyg4ometry import visualisation as pyg4vis

from .visualization import load_color_auxvals_recursive

log = logging.getLogger(__name__)


def visualize(registry: g4.Registry, scenes: dict | None = None, points=None) -> None:
    """Open a VTK-based viewer for the geometry and scene definition.

    Parameters
    ----------
    registry
        registry instance containing the geometry to view.
    scenes
        loaded :ref:`scene definition file <scene-file>`. note that the `fine_mesh`
        key is ignored and has to be set before loading/constructing the geometry.
    points
        show points, additionally to the points defined in the scene config.
    """
    if scenes is None:
        scenes = {}

    try:
        v = pyg4vis.VtkViewerColouredNew(defaultCutters=False, axisCubeWidget=False)
    except TypeError:
        v = pyg4vis.VtkViewerColouredNew()
    v.addLogicalVolume(registry.worldVolume)

    v.pygeom_scenes = scenes

    load_color_auxvals_recursive(registry.worldVolume)
    registry.worldVolume.pygeom_color_rgba = False  # hide the wireframe of the world.
    _color_recursive(registry.worldVolume, v, scenes.get("color_overrides", {}))

    # add clippers.
    clippers = scenes.get("clipper", [])
    clippers_to_remove = []
    if len(clippers) > 1:
        msg = "only one clipper can be set at the same time."
        raise ValueError(msg)
    for clip in clippers:
        clippers_to_remove = clip.get("close_cuts_remove", [])
        v.addClipper(
            clip["origin"],
            clip["normal"],
            bClipperCloseCuts=clip.get("close_cuts", False),
        )

    v.buildPipelinesAppend()

    # implement partial clipping. this can be quite confusing, as this can only remove
    # based on shared VisOptions, and not by volume name. if a volume shares their
    # VisOptions with a volume in clippers_to_remove, the closing planes of both will
    # be removed.
    if len(clippers_to_remove) > 0:
        clip_remove_regex = f"({'|'.join(clippers_to_remove)})$"
        vo_to_remove_clip = {
            str(vo)
            for k, vol in v.instanceVisOptions.items()
            if re.match(clip_remove_regex, k)
            for vo in vol
        }
        for vo in vo_to_remove_clip:
            v.ren.RemoveActor(v.actors[f"{vo}_clipper"])

    v.addAxes(length=5000)
    v.axes[0].SetVisibility(False)  # hide axes by default.
    v.addAxesWidget()
    v.axesWidget.SetEnabled(False)  # hide axes widget by default.

    if points is not None:
        _add_points(v, points)

    for scene_points in scenes.get("points", []):
        points_array = _load_points(
            scene_points["file"],
            scene_points["table"],
            scene_points.get("columns", ["xloc", "yloc", "zloc"]),
            scene_points.get("n_rows", None),
            scene_points.get("evtid", None),
        )
        _add_points(
            v,
            points_array,
            scene_points.get("color", (1, 1, 0, 1)),
            scene_points.get("size", 5),
        )

    # add light and shadow.
    if "light" in scenes:
        light = scenes["light"]
        _add_light_and_shadow(
            v, light_pos=light.get("pos"), shadow=light.get("shadow", True)
        )

    # if this option is set, do not use the interactor style (interactive=False below),
    # and directly trigger an export below.
    export_and_exit = scenes.get("export_and_exit")

    if not export_and_exit:
        # override the interactor style.
        v.interactorStyle = _KeyboardInteractor(v.ren, v.iren, v, scenes)
        v.interactorStyle.SetDefaultRenderer(v.ren)
        v.iren.SetInteractorStyle(v.interactorStyle)

    # set some defaults
    _set_camera_scene(v, scenes.get("default"))

    if "window_size" in scenes:
        v.renWin.SetSize(*scenes.get("window_size"))

    if export_and_exit:
        # force a headless mode.
        graphics_factory = vtk.vtkGraphicsFactory()
        graphics_factory.SetOffScreenOnlyMode(1)
        graphics_factory.SetUseMesaClasses(1)

        v.renWin.SetOffScreenRendering(1)

    v.view(interactive=export_and_exit is None)

    if export_and_exit:
        # export and immediately close.
        _export_png(v, file_name=export_and_exit)
        v.renWin.Finalize()
        v.iren.TerminateApp()
        del v.renWin


class _KeyboardInteractor(vtk.vtkInteractorStyleTrackballCamera):
    def __init__(self, renderer, iren, vtkviewer, scenes):
        self.AddObserver("KeyPressEvent", self.keypress)

        self.ren = renderer
        self.iren = iren
        self.vtkviewer = vtkviewer
        self.scenes = scenes

    def keypress(self, _obj, _event):
        # predefined: _e_xit

        key = self.iren.GetKeySym()
        if key == "a":  # toggle _a_xes
            ax = self.vtkviewer.axes[0]
            ax.SetVisibility(not ax.GetVisibility())
            axw = self.vtkviewer.axesWidget
            axw.SetEnabled(not axw.GetEnabled())
            axw.SetInteractive(not axw.GetEnabled())
            self.ren.GetRenderWindow().Render()

        if (
            key == "v"  # toggle _v_ertices
            and hasattr(self.vtkviewer, "points")
            and self.vtkviewer.points is not None
        ):
            pn = self.vtkviewer.points
            pn.SetVisibility(not pn.GetVisibility())
            self.ren.GetRenderWindow().Render()

        if key == "u":  # _u_p
            _set_camera(self.vtkviewer, up=(0, 0, 1), pos=(-20000, 0, 0))

        if key == "t":  # _t_op
            _set_camera(self.vtkviewer, up=(1, 0, 0), pos=(0, 0, +20000))

        if key == "b":  # _b_ottom
            _set_camera(self.vtkviewer, up=(-1, 0, 0), pos=(0, 0, -20000))

        if key == "p":  # _p_arralel projection
            cam = self.ren.GetActiveCamera()
            _set_camera(self.vtkviewer, parallel=not cam.GetParallelProjection())

        for sc_index, sc in enumerate(self.scenes.get("scenes", [])):
            if key == f"F{sc_index + 1}":
                _set_camera_scene(self.vtkviewer, sc)

        if key == "Home":
            _set_camera_scene(self.vtkviewer, self.scenes.get("default"))

        if key == "s":  # _s_ave
            _export_png(self.vtkviewer)

        if key == "i":  # dump camera _i_nfo
            cam = self.ren.GetActiveCamera()
            print(f"- focus: {list(cam.GetFocalPoint())}")  # noqa: T201
            print(f"  up: {list(cam.GetViewUp())}")  # noqa: T201
            print(f"  camera: {list(cam.GetPosition())}")  # noqa: T201
            if cam.GetParallelProjection():
                print(f"  parallel: {cam.GetParallelScale()}")  # noqa: T201

        if key == "plus":
            _set_camera(self.vtkviewer, dolly=1.1)
        if key == "minus":
            _set_camera(self.vtkviewer, dolly=0.9)


def _set_camera(
    v: pyg4vis.VtkViewerColouredNew,
    focus: tuple[float, float, float] | None = None,
    up: tuple[float, float, float] | None = None,
    pos: tuple[float, float, float] | None = None,
    dolly: float | None = None,
    parallel: bool | int | None = None,
) -> None:
    cam = v.ren.GetActiveCamera()
    if focus is not None:
        cam.SetFocalPoint(*focus)
    if up is not None:
        cam.SetViewUp(*up)
    if pos is not None:
        cam.SetPosition(*pos)
    if dolly is not None:
        if cam.GetParallelProjection():
            cam.SetParallelScale(1 / dolly * cam.GetParallelScale())
        else:
            cam.Dolly(dolly)
    if parallel is not None:
        cam.SetParallelProjection(int(parallel) > 0)
        if cam.GetParallelScale() == 1.0:
            # still at initial value, set to something more useful.
            cam.SetParallelScale(2000)
        if not isinstance(parallel, bool):
            cam.SetParallelScale(int(parallel))

    v.ren.ResetCameraClippingRange()
    v.ren.GetRenderWindow().Render()


def _set_camera_scene(v: pyg4vis.VtkViewerColouredNew, sc: dict) -> None:
    if sc is None:
        _set_camera(v, up=(1, 0, 0), pos=(0, 0, +20000))
    else:
        _set_camera(
            v,
            up=sc.get("up"),
            pos=sc.get("camera"),
            focus=sc.get("focus"),
            parallel=sc.get("parallel", False),
        )


def _export_png(v: pyg4vis.VtkViewerColouredNew, file_name: str = "scene.png") -> None:
    larger = vtk.vtkRenderLargeImage()
    larger.SetInput(v.ren)
    larger.SetMagnification(v.pygeom_scenes.get("export_scale", 1))

    # get a non-colliding file name.
    p = Path(file_name)
    stem = p.stem
    idx = 0
    while p.exists():
        p = p.with_stem(f"{stem}_{idx}")
        idx += 1
        if idx > 1000:
            msg = "could not find file name"
            raise ValueError(msg)

    png = vtk.vtkPNGWriter()
    png.SetFileName(str(p.absolute()))
    png.SetInputConnection(larger.GetOutputPort())
    png.Write()


def _add_points(v, points, color=(1, 1, 0, 1), size=5) -> None:
    # create vtkPolyData from points.
    vp = vtk.vtkPoints()
    ca = vtk.vtkCellArray()
    pd = vtk.vtkPolyData()

    for t in points:
        p = vp.InsertNextPoint(*t)
        ca.InsertNextCell(1)
        ca.InsertCellPoint(p)

    pd.SetPoints(vp)
    pd.SetVerts(ca)

    # add points to renderer.
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputData(pd)
    mapper.ScalarVisibilityOff()

    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    actor.GetProperty().SetColor(*color[0:3])
    actor.GetProperty().SetPointSize(size)
    actor.GetProperty().SetOpacity(color[3])
    actor.GetProperty().SetRenderPointsAsSpheres(True)

    v.ren.AddActor(actor)
    v.points = actor


def _load_points(
    lh5_file: str,
    point_table: str,
    columns: list[str],
    n_rows: int | None,
    evtid: int | None = None,
) -> np.ndarray:
    import pint
    from lgdo import lh5

    u = pint.get_application_registry()

    log.info(
        "loading table %s (with columns %s) from file %s",
        point_table,
        str(columns),
        lh5_file,
    )
    extra_kwargs = {}
    if n_rows is not None:
        extra_kwargs["n_rows"] = n_rows
    point_table = lh5.read(point_table, lh5_file, **extra_kwargs)

    cols_to_load = list(columns)
    if evtid is not None:
        cols_to_load.append("evtid")
    tbl = ak.Array(
        {col: point_table[col].view_as("ak", with_units=True) for col in cols_to_load}
    )

    # the points need to be in mm.
    cols = []
    for c in columns:
        col = tbl[c]
        col_units = ak.parameters(col).get(
            "units", ak.parameters(tbl).get("units", None)
        )
        factor = (u(col_units) / u("mm")).to("dimensionless").m
        if evtid is not None:
            col = col[tbl.evtid == evtid]
        if col.ndim > 1:
            col = ak.flatten(col)
        cols.append(col.to_numpy() * factor)

    return np.array(cols).T


def _color_override_matches(overrides: dict, name: str):
    for pattern, color in overrides.items():
        if re.match(f"{pattern}$", name):
            return color
    return None


def _color_recursive(
    lv: g4.LogicalVolume, viewer: pyg4vis.ViewerBase, overrides: dict, level: int = 0
) -> None:
    if level == 0:
        # first, make sure that we have independent VisOption instances everywhere.
        default_vo = viewer.getDefaultVisOptions()
        for vol in viewer.instanceVisOptions:
            viewer.instanceVisOptions[vol] = [
                copy.copy(vo) if vo is default_vo else vo
                for vo in viewer.instanceVisOptions[vol]
            ]

    if hasattr(lv, "pygeom_colour_rgba"):
        log.warning(
            "pygeom_colour_rgba on volume %s not supported, use use pygeom_color_rgba instead.",
            lv.name,
        )

    color_override = _color_override_matches(overrides, lv.name)
    if hasattr(lv, "pygeom_color_rgba") or color_override is not None:
        color_rgba = lv.pygeom_color_rgba if hasattr(lv, "pygeom_color_rgba") else None
        color_rgba = color_override if color_override is not None else color_rgba
        assert color_rgba is not None

        for vis in viewer.instanceVisOptions[lv.name]:
            if color_rgba is False:
                vis.alpha = 0
                vis.visible = False
            else:
                vis.colour = color_rgba[0:3]
                vis.alpha = color_rgba[3]
                vis.visible = vis.alpha > 0

    for pv in lv.daughterVolumes:
        if pv.type == "placement":
            _color_recursive(pv.logicalVolume, viewer, overrides, level + 1)


def _add_light_and_shadow(
    v: pyg4vis.VtkViewerColouredNew,
    light_pos: tuple[float, float, float],
    shadow: bool,
) -> None:
    lc = vtk.vtkNamedColors()
    lc.SetColor("LightColor", [255, 255, 251, 255])
    light = vtk.vtkLight()
    light.SetFocalPoint(0, 0, 0)
    light.SetPosition(*light_pos)
    light.SetIntensity(1)
    light.SetColor(lc.GetColor3d("LightColor"))
    v.ren.AddLight(light)

    if shadow:
        passes = vtk.vtkRenderPassCollection()
        shadows = vtk.vtkShadowMapPass()
        passes.AddItem(shadows.GetShadowMapBakerPass())
        passes.AddItem(shadows)
        passes.AddItem(vtk.vtkDefaultPass())
        seq = vtk.vtkSequencePass()
        seq.SetPasses(passes)

        camera_pass = vtk.vtkCameraPass()
        camera_pass.SetDelegatePass(seq)
        v.ren.SetPass(camera_pass)


def vis_gdml_cli(args: list[str] | None = None) -> None:
    parser = argparse.ArgumentParser(
        prog="legend-pygeom-vis",
        description="%(prog)s command line interface",
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="""Increase the program verbosity""",
    )
    parser.add_argument(
        "--debug",
        "-d",
        action="store_true",
        help="""Increase the program verbosity to maximum""",
    )
    parser.add_argument(
        "--fine",
        action="store_true",
        help="""use finer meshing settings""",
    )
    parser.add_argument(
        "--scene",
        "-s",
        help="""scene definition file.""",
    )
    parser.add_argument(
        "--add-points",
        help="""load points from LH5 file""",
    )
    parser.add_argument(
        "--add-points-columns",
        default="vtx:xloc,yloc,zloc",
        help="""columns in the point file. default: %(default)s""",
    )

    parser.add_argument(
        "filename",
        help="""GDML file to visualize.""",
    )

    args = parser.parse_args(args)

    logging.basicConfig()
    if args.verbose:
        logging.getLogger("pygeomtools").setLevel(logging.DEBUG)
    if args.debug:
        logging.root.setLevel(logging.DEBUG)

    scene = {}
    if args.scene:
        scene = load_dict(args.scene)

    if scene.get("fine_mesh", args.fine):
        meshconfig.setGlobalMeshSliceAndStack(100)

    points = None
    if args.add_points:
        table_parts = [c.strip() for c in args.add_points_columns.split(":")]
        point_table = table_parts[0]
        point_columns = [c.strip() for c in table_parts[1].split(",")]
        if len(table_parts) != 2 or len(point_columns) != 3:
            msg = "invalid parameter for points"
            raise ValueError(msg)

        points = _load_points(args.add_points, point_table, point_columns, None, None)

    log.info("loading GDML geometry from %s", args.filename)
    registry = gdml.Reader(args.filename).getRegistry()

    log.info("visualizing...")
    visualize(registry, scene, points)



from __future__ import annotations

import pyg4ometry.geant4 as g4
import pytest

from pygeomtools.geometry import check_materials


@pytest.fixture
def dummy_mat():
    reg = g4.Registry()
    mat = g4.Material(name="m", density=1, number_of_components=2, registry=reg)
    e1 = g4.ElementSimple(name="E1", symbol="E1", Z=1, A=1, registry=reg)
    e2 = g4.ElementSimple(name="E2", symbol="E2", Z=1, A=2, registry=reg)
    return reg, mat, e1, e2


def test_material_normal(dummy_mat):
    reg, mat, e1, e2 = dummy_mat
    mat.add_element_massfraction(e1, massfraction=0.2)
    mat.add_element_massfraction(e2, massfraction=0.8)
    check_materials(reg)


def test_material_wrong_sum(dummy_mat):
    reg, mat, e1, e2 = dummy_mat
    mat.add_element_massfraction(e1, massfraction=0.2)
    mat.add_element_massfraction(e2, massfraction=0.7)

    with pytest.warns(RuntimeWarning, match="massfraction"):
        check_materials(reg)


def test_material_duplicate_element(dummy_mat):
    reg, mat, e1, e2 = dummy_mat
    mat.add_element_massfraction(e1, massfraction=0.2)
    mat.add_element_massfraction(e1, massfraction=0.8)

    with pytest.warns(RuntimeWarning, match="duplicate elements"):
        check_materials(reg)


def test_material_component_mixture(dummy_mat):
    reg, mat, e1, e2 = dummy_mat
    mat.add_element_massfraction(e1, massfraction=0.2)
    mat.add_element_natoms(e2, natoms=2)

    with pytest.warns(RuntimeWarning) as record:
        check_materials(reg)
    assert len(record) == 2
    assert str(record[0].message) == "Material m with invalid massfraction sum 0.200"
    assert str(record[1].message) == "Material m with component type mixture"



from __future__ import annotations

import pyg4ometry
import pytest
from pyg4ometry import geant4 as g4


def test_vis_macro(tmp_path):
    from pygeomtools import visualization, write_pygeom

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    lar_mat = g4.MaterialPredefined("G4_lAr")
    scint = g4.solid.Box("scint", 0.5, 1, 1, registry, "m")
    scint1 = g4.LogicalVolume(scint, lar_mat, "scint1", registry)
    scint2 = g4.LogicalVolume(scint, lar_mat, "scint2", registry)
    g4.PhysicalVolume([0, 0, 0], [-255, 0, 0], scint1, "scint1", world_lv, registry)
    g4.PhysicalVolume([0, 0, 0], [+255, 0, 0], scint2, "scint2", world_lv, registry)
    scint1.pygeom_color_rgba = False
    scint2.pygeom_color_rgba = [1, 0, 1, 0.5]

    det = g4.solid.Box("det", 0.1, 0.5, 0.5, registry, "m")
    det = g4.LogicalVolume(det, g4.MaterialPredefined("G4_Ge"), "det", registry)
    g4.PhysicalVolume([0, 0, 0], [0, 0, 0], det, "det1", scint1, registry)
    det.pygeom_color_rgba = [0.3, 0, 1, 0.5]

    write_pygeom(registry, tmp_path / "geometry-vis.gdml")

    # test read again
    registry = pyg4ometry.gdml.Reader(tmp_path / "geometry-vis.gdml").getRegistry()

    visualization.load_color_auxvals_recursive(registry.worldVolume)

    assert not registry.logicalVolumeDict["scint1"].pygeom_color_rgba
    assert registry.logicalVolumeDict["scint2"].pygeom_color_rgba == [1, 0, 1, 0.5]
    assert registry.logicalVolumeDict["det"].pygeom_color_rgba == [0.3, 0, 1, 0.5]

    # test macro generation.
    visualization.generate_color_macro(registry, tmp_path / "color.mac")
    with (tmp_path / "color.mac").open() as f:
        generated_macro = f.read()

    expected_macro = """
/vis/geometry/set/forceSolid scint1
/vis/geometry/set/visibility scint1 -1 false
/vis/geometry/set/forceSolid det
/vis/geometry/set/colour det 0 0.3 0.0 1.0 0.5
/vis/geometry/set/forceSolid scint2
/vis/geometry/set/colour scint2 0 1.0 0.0 1.0 0.5
"""
    assert generated_macro.strip() == expected_macro.strip()


def test_double_write(tmp_path):
    from pygeomtools import visualization, write_pygeom

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    lar_mat = g4.MaterialPredefined("G4_lAr")
    scint = g4.solid.Box("scint", 0.5, 1, 1, registry, "m")
    scint1 = g4.LogicalVolume(scint, lar_mat, "scint1", registry)
    scint2 = g4.LogicalVolume(scint, lar_mat, "scint2", registry)
    g4.PhysicalVolume([0, 0, 0], [-255, 0, 0], scint1, "scint1", world_lv, registry)
    g4.PhysicalVolume([0, 0, 0], [+255, 0, 0], scint2, "scint2", world_lv, registry)
    scint1.pygeom_color_rgba = False
    scint2.pygeom_color_rgba = [1, 0, 1, 0.5]

    write_pygeom(registry, tmp_path / "geometry-vis.gdml")

    # simulate a second write
    registry.userInfo = []
    write_pygeom(registry, tmp_path / "geometry-vis.gdml")

    # test read again
    registry = pyg4ometry.gdml.Reader(tmp_path / "geometry-vis.gdml").getRegistry()

    visualization.load_color_auxvals_recursive(registry.worldVolume)


def test_vis_typo(tmp_path):
    from pygeomtools import write_pygeom

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    scint = g4.solid.Box("scint", 0.5, 1, 1, registry, "m")
    scint1 = g4.LogicalVolume(scint, g4.MaterialPredefined("G4_lAr"), "scint", registry)
    g4.PhysicalVolume([0, 0, 0], [-255, 0, 0], scint1, "scint", world_lv, registry)
    scint1.pygeom_colour_rgba = False

    with pytest.raises(RuntimeError, match="use use pygeom_color_rgba instead"):
        write_pygeom(registry, tmp_path / "geometry-vis2.gdml")



from __future__ import annotations

import pyg4ometry
import pytest
from dbetto import AttrsDict
from pyg4ometry import geant4 as g4


def test_detector_info(tmp_path):
    from pygeomtools import RemageDetectorInfo, detectors, write_pygeom

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    lar_mat = g4.MaterialPredefined("G4_lAr")
    scint = g4.solid.Box("scint", 0.5, 1, 1, registry, "m")
    scint1 = g4.LogicalVolume(scint, lar_mat, "scint1", registry)
    scint2 = g4.LogicalVolume(scint, lar_mat, "scint2", registry)
    scint1pv = g4.PhysicalVolume(
        [0, 0, 0], [-255, 0, 0], scint1, "scint1", world_lv, registry
    )
    scint1pv.set_pygeom_active_detector(RemageDetectorInfo("scintillator", 3))
    scint2pv = g4.PhysicalVolume(
        [0, 0, 0], [+255, 0, 0], scint2, "scint2", world_lv, registry
    )
    assert scint2pv.get_pygeom_active_detector() is None
    scint2pv.set_pygeom_active_detector(RemageDetectorInfo("scintillator", 3))
    assert scint2pv.pygeom_active_detector is not None
    assert scint2pv.get_pygeom_active_detector() == scint2pv.pygeom_active_detector

    det = g4.solid.Box("det", 0.1, 0.5, 0.5, registry, "m")
    det = g4.LogicalVolume(det, g4.MaterialPredefined("G4_Ge"), "det", registry)
    det1 = g4.PhysicalVolume([0, 0, 0], [0, 0, 0], det, "det1", scint1, registry)
    det1.pygeom_active_detector = RemageDetectorInfo("optical", 1, {"some": "metadata"})
    det2 = g4.PhysicalVolume([0, 0, 0], [0, 0, 0], det, "det2", scint2, registry)
    det2.pygeom_active_detector = RemageDetectorInfo(
        "germanium", 2, {"other": "other metadata"}
    )

    write_pygeom(registry, tmp_path / "geometry.gdml", ignore_duplicate_uids=True)
    detectors.generate_detector_macro(registry, tmp_path / "geometry.mac")
    expected_macro = """
/RMG/Geometry/RegisterDetector Germanium det2 2
/RMG/Geometry/RegisterDetector Optical det1 1
/RMG/Geometry/RegisterDetector Scintillator scint1 3
/RMG/Geometry/RegisterDetector Scintillator scint2 3
"""
    assert (tmp_path / "geometry.mac").read_text().strip() == expected_macro.strip()

    # test read again
    registry = pyg4ometry.gdml.Reader(tmp_path / "geometry.gdml").getRegistry()

    all_top_level_aux = [(aux.auxtype, aux.auxvalue) for aux in registry.userInfo]
    assert all_top_level_aux == [
        ("RMG_detector_meta", ""),
        ("RMG_detector", "germanium"),
        ("RMG_detector", "optical"),
        ("RMG_detector", "scintillator"),
    ]

    # test that our API functions work.
    assert detectors.get_sensvol_metadata(registry, "det2") == {
        "other": "other metadata"
    }
    det1meta = detectors.get_sensvol_metadata(registry, "det1")
    assert det1meta == {"some": "metadata"}
    assert det1meta.some == "metadata"
    assert isinstance(det1meta, AttrsDict)
    assert detectors.get_sensvol_metadata(registry, "scint1") is None
    sensvols = detectors.get_all_sensvols(registry)
    assert set(sensvols.keys()) == {"det2", "det1", "scint1", "scint2"}
    assert sensvols["scint1"].uid == 3

    # test retrieval by uid.
    assert detectors.get_sensvol_by_uid(registry, 3) == ("scint1", sensvols["scint1"])
    assert detectors.get_sensvol_by_uid(registry, 5) is None


def test_detector_typo(tmp_path):
    from pygeomtools import RemageDetectorInfo, detectors, write_pygeom

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    scint = g4.solid.Box("scint", 0.5, 1, 1, registry, "m")
    scint1 = g4.LogicalVolume(
        scint, g4.MaterialPredefined("G4_lAr"), "scint1", registry
    )
    scint1pv = g4.PhysicalVolume(
        [0, 0, 0], [-255, 0, 0], scint1, "scint1", world_lv, registry
    )
    with pytest.deprecated_call():
        scint1pv.pygeom_active_dector = RemageDetectorInfo("scintillator", 3)
    assert scint1pv.get_pygeom_active_detector() is not None
    assert scint1pv.pygeom_active_detector is not None
    with pytest.deprecated_call():
        assert scint1pv.pygeom_active_dector is not None

    write_pygeom(registry, tmp_path / "geometry_typo.gdml")

    # test read again
    registry = pyg4ometry.gdml.Reader(tmp_path / "geometry_typo.gdml").getRegistry()

    all_top_level_aux = [(aux.auxtype, aux.auxvalue) for aux in registry.userInfo]
    assert all_top_level_aux == [
        ("RMG_detector_meta", ""),
        ("RMG_detector", "scintillator"),
    ]

    sensvols = detectors.get_all_sensvols(registry)
    assert set(sensvols.keys()) == {"scint1"}
    assert sensvols["scint1"].uid == 3
    assert detectors.get_sensvol_by_uid(registry, 3) == ("scint1", sensvols["scint1"])


def test_no_detector_info(tmp_path):
    from pygeomtools import detectors, write_pygeom

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    scint = g4.solid.Box("scint", 0.5, 1, 1, registry, "m")
    scint = g4.LogicalVolume(scint, g4.MaterialPredefined("G4_lAr"), "scint1", registry)
    g4.PhysicalVolume([0, 0, 0], [-255, 0, 0], scint, "scint1", world_lv, registry)

    write_pygeom(registry, tmp_path / "geometry_no_det.gdml")
    detectors.generate_detector_macro(registry, tmp_path / "geometry_no_det.mac")
    assert (tmp_path / "geometry_no_det.mac").read_text() == ""

    # test read again
    registry = pyg4ometry.gdml.Reader(tmp_path / "geometry_no_det.gdml").getRegistry()
    assert detectors.get_sensvol_metadata(registry, "det1") is None

    all_top_level_aux = [(aux.auxtype, aux.auxvalue) for aux in registry.userInfo]
    assert all_top_level_aux == [("RMG_detector_meta", "")]
    assert registry.userInfo[0].subaux == []


def test_wrong_write(tmp_path):
    from pygeomtools import detectors

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    scint = g4.solid.Box("scint", 0.5, 1, 1, registry, "m")
    scint = g4.LogicalVolume(scint, g4.MaterialPredefined("G4_lAr"), "scint1", registry)
    g4.PhysicalVolume([0, 0, 0], [-255, 0, 0], scint, "scint1", world_lv, registry)

    w = pyg4ometry.gdml.Writer()
    w.addDetector(registry)
    w.write(str(tmp_path / "geometry_wrong_write.gdml"))

    # test read again
    registry = pyg4ometry.gdml.Reader(
        tmp_path / "geometry_wrong_write.gdml"
    ).getRegistry()
    with pytest.raises(RuntimeError):
        detectors.get_sensvol_metadata(registry, "det1")
    with pytest.raises(RuntimeError):
        detectors.get_all_sensvols(registry)
    with pytest.raises(RuntimeError):
        detectors.get_sensvol_by_uid(registry, 5)

    all_top_level_aux = [(aux.auxtype, aux.auxvalue) for aux in registry.userInfo]
    assert all_top_level_aux == []


def test_duplicate_uid():
    from pygeomtools import RemageDetectorInfo, detectors

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    det = g4.solid.Box("det", 0.1, 0.5, 0.5, registry, "m")
    det = g4.LogicalVolume(det, g4.MaterialPredefined("G4_Ge"), "det", registry)
    det1 = g4.PhysicalVolume([0, 0, 0], [-255, 0, 0], det, "det1", world_lv, registry)
    det1.pygeom_active_detector = RemageDetectorInfo("optical", 1, {"some": "metadata"})
    det2 = g4.PhysicalVolume([0, 0, 0], [+255, 0, 0], det, "det2", world_lv, registry)
    det2.pygeom_active_detector = RemageDetectorInfo(
        "germanium", 1, {"other": "other metadata"}
    )

    with pytest.raises(RuntimeError):
        assert not detectors.check_detector_uniqueness(registry)


def test_unknown_type():
    from pygeomtools import RemageDetectorInfo, detectors

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    det = g4.solid.Box("det", 0.1, 0.5, 0.5, registry, "m")
    det = g4.LogicalVolume(det, g4.MaterialPredefined("G4_Ge"), "det", registry)
    det1 = g4.PhysicalVolume([0, 0, 0], [-255, 0, 0], det, "det1", world_lv, registry)
    det1.pygeom_active_detector = RemageDetectorInfo("abc", 1)

    with pytest.raises(RuntimeError, match="unknown detector_type"):
        detectors.write_detector_auxvals(registry)


def test_wrong_uid():
    from pygeomtools import RemageDetectorInfo, detectors

    registry = g4.Registry()
    world = g4.solid.Box("world", 2, 2, 2, registry, "m")
    world_lv = g4.LogicalVolume(
        world, g4.MaterialPredefined("G4_Galactic"), "world", registry
    )
    registry.setWorld(world_lv)

    det = g4.solid.Box("det", 0.1, 0.5, 0.5, registry, "m")
    det = g4.LogicalVolume(det, g4.MaterialPredefined("G4_Ge"), "det", registry)
    det1 = g4.PhysicalVolume([0, 0, 0], [-255, 0, 0], det, "det1", world_lv, registry)
    det1.pygeom_active_detector = RemageDetectorInfo("optical", "x")

    with pytest.raises(ValueError):
        detectors.write_detector_auxvals(registry)



"""Assignment of sensitive detectors to physical volumes, for use in ``remage``."""

from __future__ import annotations

import json
import logging
from collections.abc import Generator
from dataclasses import dataclass
from itertools import groupby
from pathlib import Path
from typing import Literal, get_args, get_type_hints

import pyg4ometry.geant4 as g4
from dbetto import AttrsDict
from pyg4ometry.gdml.Defines import Auxiliary

log = logging.getLogger(__name__)

AUXKEY_DETMETA = "RMG_detector_meta"
AUXKEY_DET = "RMG_detector"


@dataclass
class RemageDetectorInfo:
    detector_type: Literal["optical", "germanium", "scintillator"]
    """``remage`` detector type."""

    uid: int
    """``remage`` detector UID."""

    metadata: object | None = None
    """Attach arbitrary metadata to this sensitive volume. This will be written to GDML as JSON.

    See also
    ========
    .get_sensvol_metadata
    """


def walk_detectors(
    pv: g4.PhysicalVolume | g4.LogicalVolume | g4.Registry,
) -> Generator[tuple[g4.PhysicalVolume, RemageDetectorInfo], None, None]:
    """Iterate over all physical volumes that have a :class:`RemageDetectorInfo` attached.

    Important
    ---------
    this only returns instances previously set via
    :meth:`get_pygeom_active_detector() <pyg4ometry.geant4.PhysicalVolume.get_pygeom_active_detector>`,
    not data loaded from GDML. Use :meth:`get_all_sensvols` instead for that use case.
    """

    if isinstance(pv, g4.PhysicalVolume):
        det = pv.get_pygeom_active_detector()
        if det is not None:
            assert isinstance(det, RemageDetectorInfo)
            yield pv, det
    if isinstance(pv, g4.LogicalVolume):
        next_v = pv
    if isinstance(pv, g4.PhysicalVolume):
        next_v = pv.logicalVolume
    elif isinstance(pv, g4.Registry):
        next_v = pv.worldVolume
    else:
        msg = "invalid type encountered in walk_detectors volume tree"
        raise TypeError(msg)

    for dv in next_v.daughterVolumes:
        if dv.type == "placement":
            yield from walk_detectors(dv)


def generate_detector_macro(registry: g4.Registry, filename: str) -> None:
    """Create a Geant4 macro file containing the defined active detector volumes for use in remage."""
    if _get_rmg_detector_aux(registry, raise_on_missing=False) is not None:
        sensvols = get_all_sensvols(registry)
    else:
        sensvols = {pv.name: det for pv, det in walk_detectors(registry)}

    macro_lines = {}
    for pv, det in sensvols.items():
        if pv in macro_lines:
            continue
        mac = f"/RMG/Geometry/RegisterDetector {det.detector_type.title()} {pv} {det.uid}\n"
        macro_lines[pv] = mac

    macro_contents = "".join(macro_lines.values())

    with Path(filename).open("w", encoding="utf-8") as f:
        f.write(macro_contents)


def write_detector_auxvals(registry: g4.Registry) -> None:
    """Append an auxiliary structure, storing the sensitive detector volume information.

    .. note::
        see :doc:`../metadata` for a reference of the written structure.
    """
    if _get_rmg_detector_aux(registry, raise_on_missing=False) is not None:
        msg = "detector auxiliary structure already written"
        raise RuntimeError(msg)

    written_pvs = set()
    group_it = groupby(
        sorted(walk_detectors(registry), key=lambda d: d[1].detector_type),
        lambda d: d[1].detector_type,
    )

    meta_group_aux = Auxiliary(AUXKEY_DETMETA, "", registry)

    for key, group in group_it:
        if key not in get_args(get_type_hints(RemageDetectorInfo)["detector_type"]):
            msg = f"unknown detector_type {key}"
            raise RuntimeError(msg)

        group_aux = Auxiliary(AUXKEY_DET, key, registry)

        for pv, det in group:
            if pv.name in written_pvs:
                continue
            written_pvs.add(pv.name)

            group_aux.addSubAuxiliary(
                Auxiliary(pv.name, str(int(det.uid)), registry, addRegistry=False)
            )
            if det.metadata is not None:
                json_meta = json.dumps(det.metadata, sort_keys=True)
                meta_group_aux.addSubAuxiliary(
                    Auxiliary(pv.name, json_meta, registry, addRegistry=False)
                )


def check_detector_uniqueness(
    registry: g4.Registry, ignore_duplicate_uids: set[int] | None = None
) -> bool:
    """Check that each sensitive detector uid is only used once.

    Parameters
    ----------
    ignore_duplicate_uids
        a set of uids to exclude from the uniqueness check.
    """
    uids = [d[1].uid for d in walk_detectors(registry)]
    uids_to_check = set(uids) - (ignore_duplicate_uids or set())
    duplicates = [
        uid for uid in uids_to_check if len([u for u in uids if u == uid]) > 1
    ]
    if duplicates != []:
        msg = f"found duplicate detector uids {duplicates}"
        raise RuntimeError(msg)
    return duplicates == []


def _get_rmg_detector_aux(
    registry: g4.Registry, *, raise_on_missing: bool = True
) -> Auxiliary | None:
    auxs = [aux for aux in registry.userInfo if aux.auxtype == AUXKEY_DETMETA]
    if auxs == []:
        if not raise_on_missing:
            return None
        msg = f"GDML missing {AUXKEY_DETMETA} auxval (not written by legend-pygeom-tools?)"
        raise RuntimeError(msg)
    assert len(auxs) == 1
    return auxs[0]


def get_sensvol_metadata(registry: g4.Registry, name: str) -> AttrsDict | None:
    """Load metadata attached to the given sensitive volume (from GDML)."""
    meta_aux = _get_rmg_detector_aux(registry)
    meta_auxs = [aux for aux in meta_aux.subaux if aux.auxtype == name]
    if meta_auxs == []:
        return None
    assert len(meta_auxs) == 1
    return AttrsDict(json.loads(meta_auxs[0].auxvalue))


def get_all_sensvols(registry: g4.Registry) -> dict[str, RemageDetectorInfo]:
    """Load all registered sensitive detectors with their metadata (from GDML)."""
    meta_aux = _get_rmg_detector_aux(registry)
    meta_auxs = {
        aux.auxtype: AttrsDict(json.loads(aux.auxvalue)) for aux in meta_aux.subaux
    }

    detmapping = {}
    type_auxs = [aux for aux in registry.userInfo if aux.auxtype == AUXKEY_DET]
    for type_aux in type_auxs:
        for det_aux in type_aux.subaux:
            detmapping[det_aux.auxtype] = RemageDetectorInfo(
                type_aux.auxvalue, int(det_aux.auxvalue), meta_auxs.get(det_aux.auxtype)
            )

    if set(meta_auxs.keys()) - set(detmapping.keys()) != set():
        msg = "invalid GDML auxval structure (meta keys and detmapping keys differ)"
        raise RuntimeError(msg)

    return detmapping


def get_sensvol_by_uid(
    registry: g4.Registry, uid: int
) -> tuple[str, RemageDetectorInfo] | None:
    """Get the volume name and detector metadata for the detector with remage detector ID `uid`."""
    sensvols = get_all_sensvols(registry)
    found = next(filter(lambda s: s[1].uid == uid, sensvols.items()), None)
    if found is None:
        return None
    return found


def __set_pygeom_active_detector(self, det_info: RemageDetectorInfo | None) -> None:
    """Set the remage detector info on this physical volume instance."""
    if not isinstance(self, g4.PhysicalVolume):
        msg = "patched-in function called on wrong type"
        raise TypeError(msg)
    assert self.registry is not None
    if _get_rmg_detector_aux(self.registry, raise_on_missing=False) is not None:
        msg = "detector auxiliary structure already written"
        raise RuntimeError(msg)
    self.__pygeom_active_detector = det_info


def __get_pygeom_active_detector(self) -> RemageDetectorInfo | None:
    """Get the remage detector info on this physical volume instance."""
    if not isinstance(self, g4.PhysicalVolume):
        msg = "patched-in function called on wrong type"
        raise TypeError(msg)

    if hasattr(self, "__pygeom_active_detector"):
        return self.__pygeom_active_detector
    return None


def __patch_pyg4_pv():
    """monkey-patch a new function onto every PhysicalVolume instance."""
    t = g4.PhysicalVolume
    t.set_pygeom_active_detector = __set_pygeom_active_detector
    t.get_pygeom_active_detector = __get_pygeom_active_detector

    prop = property(__get_pygeom_active_detector)
    t.pygeom_active_detector = prop.setter(__set_pygeom_active_detector)

    def _wrap_warn(fn):
        def _fn(*args):
            import warnings

            warnings.warn(
                "pygeom_active_dector (typo!) is deprecated, use pygeom_active_detector instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return fn(*args)

        return _fn

    prop = property(_wrap_warn(__get_pygeom_active_detector))
    t.pygeom_active_dector = prop.setter(_wrap_warn(__set_pygeom_active_detector))


__patch_pyg4_pv()



from __future__ import annotations

# note: do not load viewer module here, as it has quite large nested imports. see lazy loading below.
from . import detectors, geometry, utils, visualization
from ._version import version as __version__
from .detectors import RemageDetectorInfo, get_all_sensvols, get_sensvol_metadata
from .write import write_pygeom

__all__ = [
    "RemageDetectorInfo",
    "__version__",
    "detectors",
    "geometry",
    "get_all_sensvols",
    "get_sensvol_metadata",
    "utils",
    "viewer",  # lazy import!
    "visualization",
    "write_pygeom",
]


# inspired by PEP 562.
def __getattr__(name: str):
    if name in __all__:
        import importlib

        return importlib.import_module(f".{name}", __name__)
    msg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(msg)



from __future__ import annotations

import logging
from collections.abc import Callable

from dbetto import AttrsDict, utils

log = logging.getLogger(__name__)


def load_dict(fname: str, ftype: str | None = None) -> dict:
    """Load a text file as a Python dict.

    .. deprecated :: 0.0.8
        Use :func:`dbetto.utils.load_dict` instead.
    """
    import warnings

    warnings.warn(
        "The load_dict function has moved to the dbetto package (https://github.com/gipert/dbetto). "
        "Please update your code, as load_dict will be removed from this package in the future.",
        DeprecationWarning,
        stacklevel=2,
    )

    return utils.load_dict(fname, ftype)


def load_dict_from_config(
    config: dict, key: str, default: Callable[[], AttrsDict]
) -> AttrsDict:
    """Helper functions to load nested data from a config file.

    * If ``key`` is in the config file
      - and it refers to a string: load a JSON/YAML file from that path.
      - and it refers to a dict: use that directly
    * else, the default value is loaded via the ``default`` callable.
    """
    m = config.get(key)
    if isinstance(m, str):
        return AttrsDict(utils.load_dict(m))
    if isinstance(m, dict):
        return AttrsDict(m)
    return default()



from __future__ import annotations

import json
from pathlib import Path

import pytest
from lgdo import Array, Table, VectorOfVectors, lh5
from pyg4ometry import gdml

from pygeomtools import viewer


@pytest.fixture
def points_file(tmp_path):
    file = tmp_path / "points.lh5"

    data = {}
    data["evtid"] = Array([0, 1])

    attrs = {"units": "m"}
    data["xloc"] = VectorOfVectors([[0.1, 0.2], [-0.3, 0.4, -0.5]], attrs=attrs)
    data["yloc"] = VectorOfVectors([[0.1, 0.2], [-0.3, 0.4, -0.5]], attrs=attrs)
    data["zloc"] = VectorOfVectors([[0.1, 0.2], [-0.3, 0.4, -0.5]], attrs=attrs)

    tab = Table(data)
    lh5.write(tab, "stp/test", file, wo_mode="of")

    return str(file)


def test_viewer(tmp_path, points_file):
    registry = gdml.Reader(Path(__file__).parent / "geometry.gdml").getRegistry()

    output_file = tmp_path / "test_viewer.png"
    output_file.unlink(missing_ok=True)

    vis_scene = {
        "window_size": [400, 700],
        "color_overrides": {
            "World": [1, 0, 0, 0.1],
        },
        "light": {
            "pos": [100, 100, 100],
            "shadow": True,
        },
        "export_scale": 2,
        "export_and_exit": output_file,
    }
    viewer.visualize(registry, vis_scene)

    # test adding a clipper.
    output_file = tmp_path / "test_viewer_clipper.png"
    output_file.unlink(missing_ok=True)

    vis_scene["clipper"] = [
        {"origin": [0, 0, 0], "normal": [0, 1, 0], "close_cuts": True}
    ]
    vis_scene["export_and_exit"] = output_file
    viewer.visualize(registry, vis_scene)

    # test showing points.
    output_file = tmp_path / "test_viewer_points.png"
    output_file.unlink(missing_ok=True)

    vis_scene["points"] = [
        {"file": points_file, "table": "stp/test", "size": 20, "color": [1, 0, 0, 1]}
    ]
    vis_scene["export_and_exit"] = output_file
    viewer.visualize(registry, vis_scene)


def test_viewer_cli(tmp_path, points_file):
    geom = Path(__file__).parent / "geometry.gdml"

    output_file = tmp_path / "test_viewer_cli.png"
    output_file.unlink(missing_ok=True)

    tmp_scene = tmp_path / "scene.yml"
    vis_scene = {
        "window_size": [400, 700],
        "export_scale": 2,
        "export_and_exit": str(output_file),
    }
    tmp_scene.write_text(json.dumps(vis_scene))

    viewer.vis_gdml_cli(["--scene", str(tmp_scene), str(geom)])

    # test showing points.
    output_file = tmp_path / "test_viewer_cli_points.png"
    output_file.unlink(missing_ok=True)

    vis_scene["export_and_exit"] = str(output_file)
    tmp_scene.write_text(json.dumps(vis_scene))
    viewer.vis_gdml_cli(
        [
            "--scene",
            str(tmp_scene),
            "--add-points",
            points_file,
            "--add-points-columns",
            "stp/test:xloc,yloc,zloc",
            str(geom),
        ]
    )



"""Helper functions for geometry construction."""

from __future__ import annotations

import warnings
from collections import Counter

from pyg4ometry import geant4


def check_registry_sanity(v, registry: geant4.Registry) -> None:
    """Check recursively if all children in the volume and material tree have the correct
    registry instance attached.

    Parameters
    ==========
    v
        object to recursively check to have the right registry.
    registry
        the expected registry to compare against

    Note
    ====
    This function prevents an easy-to-miss problem using pyg4ometry: If different (or no)
    registries are used inside an object structure, this might lead to unexpected results
    in GDML output.
    """
    if not isinstance(v, geant4.Registry) and v.registry is not registry:
        msg = f"found invalid registry instance on {v}"
        raise RuntimeError(msg)

    # walk the tree.
    if isinstance(v, geant4.LogicalVolume | geant4.AssemblyVolume):
        for dv in v.daughterVolumes:
            check_registry_sanity(dv, registry)
        check_registry_sanity(v.material, registry)
        check_registry_sanity(v.solid, registry)

    elif isinstance(v, geant4.PhysicalVolume):
        check_registry_sanity(v.logicalVolume, registry)

    elif isinstance(v, geant4.Registry):
        check_registry_sanity(v.worldVolume, registry)
        for s in v.surfaceDict.values():
            check_registry_sanity(s, registry)

        check_materials(registry)

    elif isinstance(v, geant4.Material):
        if hasattr(v, "components"):
            for comp in v.components:
                check_registry_sanity(comp[0], registry)
            if v not in registry.materialDict.values():
                warnings.warn(
                    f"found material {v.name} not in materialDict",
                    RuntimeWarning,
                    stacklevel=1,
                )

    elif isinstance(v, geant4.SurfaceBase):
        check_registry_sanity(v.surface_property, registry)

    elif isinstance(
        v, geant4.solid.OpticalSurface | geant4.solid.SolidBase | geant4.Element
    ):
        pass

    else:
        msg = f"invalid type {type(v)} encountered in check_registry_sanity volume tree"
        raise TypeError(msg)


def check_materials(registry: geant4.Registry) -> None:
    """Check against some common problems of materials."""
    for mat in registry.materialDict.values():
        if not hasattr(mat, "components"):
            continue

        elems = []
        mass = 0
        types = set()
        for comp in mat.components:
            elems.append(comp[0].name)
            mass += comp[1] if comp[2] == "massfraction" else 0
            types.add(comp[2])

        duplicate_elems = [elem for elem, count in Counter(elems).items() if count > 1]
        if duplicate_elems != []:
            warnings.warn(
                f"Material {mat.name} with duplicate elements {duplicate_elems}",
                RuntimeWarning,
                stacklevel=1,
            )

        if float(mass) not in (0.0, 1.0):
            warnings.warn(
                f"Material {mat.name} with invalid massfraction sum {mass:.3f}",
                RuntimeWarning,
                stacklevel=1,
            )

        if (
            len(types) > 1
            and not isinstance(mat, geant4._Material.Element)
            and not mat.name.startswith("G4")
        ):
            warnings.warn(
                f"Material {mat.name} with component type mixture",
                RuntimeWarning,
                stacklevel=1,
            )



from __future__ import annotations

import pytest
from pyg4ometry import geant4 as g4

from pygeomtools.materials import BaseMaterialRegistry, cached_property


class DummyMaterials:
    @cached_property
    def x(self) -> str:
        return {"test"}


def test_cached_property():
    m = DummyMaterials()

    first_prop_access = m.x
    assert first_prop_access == {"test"}

    assert m.x == {"test"}
    assert m.x is not {"test"}  # noqa: F632 (just a sanity check)
    assert first_prop_access is m.x  # we should get back the same object.

    # writing and deleting is not possible.
    with pytest.raises(AttributeError):
        m.x = None
    with pytest.raises(AttributeError):
        del m.x


class DummyMaterialRegistry(BaseMaterialRegistry):
    pass


def test_material_registry():
    registry = g4.Registry()
    mat = DummyMaterialRegistry(registry)

    assert registry.materialDict == {}

    ar = mat.get_element("Ar")
    assert ar.Z == 18
    assert ar.registry is registry
    assert registry.materialDict == {"argon": ar}

    # the second access should return the same object and should not raise an error.
    ar2 = mat.get_element("Ar")
    assert ar is ar2
